/****************************************************************************/
/* program to generate 3-D Cell Suppression Problem instances               */
/* command line arguments:                                                  */
/*        nrows: number of rows of the table                                */
/*        ncols: number of columns of the table                             */
/*        nlevs: number of levels of the table                              */
/*        p:     number of primary suppressions             		    */
/*        p_l%:  lower level protection as a p_l% of the cell value         */
/*        p_u%:  upper level protection as a p_u% of the cell value         */
/*        type:  type of generation (explained below)                       */
/*        seed:  integer initial seed for random numbers    		    */
/*                                                                          */
/* The structure of the generated table depends of parameter "type":        */
/*                                                                          */
/* Type = 1: (similar to Kelly-Golden-Assad generator)                      */
/*    The generator produces a positive integer table.         		    */
/*    Values of the cells are randomly obtained from an integer uniform     */
/*    distribution {1..1000} with probability 0.8 and are 0 with            */
/*    probability 0.2. Location of primary (internal) cells is randomly     */
/*    obtained.                                                             */
/*                                    		                            */
/* Type = 2: (similar to generator 1 of Fischetti-Salazar Math.Prog. paper) */
/*    The generator produces a positive integer table.          	    */
/*    Values of the p cells are randomly obtained from an integer uniform   */
/*    distribution {1..4}. Location of primary (internal) cells is randomly */
/*    obtained. Values of the remaining cells are randomly obtained from    */
/*    an integer uniform distribution {0,5..500}.                  	    */
/****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/* typedefs */
typedef struct{
  int m; /* number of rows */
  int n; /* number of columns */
  int l; /* numbe of levels */
  int *v; /* (m+1)*(n+1)*(l+1) table values 
	     (marginals included, from 0..m to 0..n to 0..l)*/
  int p; /* number of primary suppressions */
  int *p_m; /* rows of the primary suppressions */
  int *p_n; /* columns of the primary suppressions */
  int *p_l; /* levels of the primary suppressions */
  int *p_o; /* merit order of the primary suppresssions */
  int *p_lpl; /* lower protection limit for each primary suppression */
  int *p_upl; /* upper protection limit for each primary suppression */
} TABLE3D;


/* defines */
#define TRUE  (char)1
#define FALSE (char)0
#define ALLOC(n,type) (type *)malloc((n)*sizeof(type))


/* function prototypes */
void allocate_table(TABLE3D **ptab, int ncols, int nrows, int nlevs, int p);
int get_cell3D(TABLE3D *tab, int row, int column, int level);
void put_cell3D(TABLE3D *tab, int row, int column, int level, int value);
void generate_table_1(TABLE3D *tab, double pl, double pu, long seed);
void generate_table_2(TABLE3D *tab, double pl, double pu, long seed);
void compute_marginals(TABLE3D *tab);
void write_table_AMPL_format(TABLE3D *tab,double pl,double pu,int type,int seed);
int int_uniform(int min, int max);

void main(int argc, char *argv[])
{
  
  int nrows, ncols, nlevs, p, type;
  long seed;
  TABLE3D *tab;
  double pl,pu;
  
  /* check command line arguments */

  if (argc != 9) {
    fprintf(stderr,"usage: gen_CSP_instance nrows ncols nlevs p p_l%% p_u%% type seed\n");
    exit(-1);
  }
  
  sscanf(argv[1],"%d",&nrows);
  if (nrows < 1) {
    fprintf(stderr,"error: nrows must be greater than 0 \n");
    exit(-1);
  }
  sscanf(argv[2],"%d",&ncols);
  if (ncols < 1) {
    fprintf(stderr,"error: ncols must be greater than 0 \n");
    exit(-1);
  }
  sscanf(argv[3],"%d",&nlevs);
  if (nlevs < 1) {
    fprintf(stderr,"error: nlevs must be greater than 0 \n");
    exit(-1);
  }
  sscanf(argv[4],"%d",&p);
  if (p > ncols*nrows*nlevs) {
    fprintf(stderr,"error: p must be less than ncols*nrows*nlevs \n");
    exit(-1);
  }
  sscanf(argv[5],"%lf",&pl);
  if (pl <= 0.00000 || pl > 100.0) {
    fprintf(stderr,"error: p_l%% must be in (0,100] \n");
    exit(-1);
  }

  sscanf(argv[6],"%lf",&pu);
  if (pu <= 0.00000 || pu > 100.0) {
    fprintf(stderr,"error: p_u%% must be in (0,100] \n");
    exit(-1);
  }

  sscanf(argv[7],"%d",&type);
  if (type >2 || type <= 0) {
    fprintf(stderr,"error: type must be in 1..2 \n");
    exit(-1);
  }

  sscanf(argv[8],"%d",&seed);

  /* allocate table */
  allocate_table(&tab,ncols,nrows,nlevs,p);

  if (type == 1) 
    /* generate table */
    generate_table_1(tab,pl,pu,seed);
  else if (type == 2)
     generate_table_2(tab,pl,pu,seed);
  
  /* write_table in AMPL format */
  write_table_AMPL_format(tab,pl,pu,type,seed);
}

void allocate_table(TABLE3D **ptab, int ncols, int nrows, int nlevs, int p)
{
  (*ptab)= new TABLE3D;
  (*ptab)->m= nrows;
  (*ptab)->n= ncols;
  (*ptab)->l= nlevs;
  (*ptab)->p= p;
  (*ptab)->v= new int[(nrows+1)*(ncols+1)*(nlevs+1)];
  (*ptab)->p_m= new int[p];
  (*ptab)->p_n= new int[p];
  (*ptab)->p_l= new int[p];
  (*ptab)->p_o= new int[p];  
  (*ptab)->p_lpl= new int[p];  
  (*ptab)->p_upl= new int[p];;
}

int get_cell3D(TABLE3D *tab, int row, int column, int level)
{
  return(tab->v[level*(tab->n+1)*(tab->m+1)+row*(tab->n+1)+column]);
}

void put_cell3D(TABLE3D *tab, int row, int column, int level, int value)
{
  tab->v[level*(tab->n+1)*(tab->m+1)+row*(tab->n+1)+column]= value;
}

void compute_marginals(TABLE3D *tab)
{
  int i,j,k,sum,tot;

  /* marginals obtained by adding one dimension */
  tot= 0;
  for(i=0;i<tab->m;i++) {
    for(j=0;j<tab->n;j++) {
      for(sum=0,k=0;k<tab->l;k++) sum += get_cell3D(tab,i,j,k);
      put_cell3D(tab,i,j,tab->l,sum);
    }
  }
  for(i=0;i<tab->m;i++) {
    for(k=0;k<tab->l;k++) {
      for(sum=0,j=0;j<tab->n;j++) sum += get_cell3D(tab,i,j,k);
      put_cell3D(tab,i,tab->n,k,sum);
    }
  }
  for(j=0;j<tab->n;j++) {
    for(k=0;k<tab->l;k++) {
      for(sum=0,i=0;i<tab->m;i++) sum += get_cell3D(tab,i,j,k);
      put_cell3D(tab,tab->m,j,k,sum);
      tot += sum;
    }
  }

  /* marginals obtained by adding two dimensions */
  for(i=0;i<tab->m;i++) {
    for(sum=0,k=0;k<tab->l;k++) sum += get_cell3D(tab,i,tab->n,k);
    put_cell3D(tab,i,tab->n,tab->l,sum);
  }
  for(j=0;j<tab->n;j++) {
    for(sum=0,k=0;k<tab->l;k++) sum += get_cell3D(tab,tab->m,j,k);
    put_cell3D(tab,tab->m,j,tab->l,sum);
  }
  tot= 0;
  for(k=0;k<tab->l;k++) {
    for(sum=0,i=0;i<tab->m;i++) sum += get_cell3D(tab,i,tab->n,k);
    put_cell3D(tab,tab->m,tab->n,k,sum);
    tot += sum;
  }

  /* marginals obtained by adding three dimensions */
  put_cell3D(tab,tab->m,tab->n,tab->l,tot);
}

void generate_table_1(TABLE3D *tab, double pl, double pu, long seed)
{
  int i,j,k,r,c,l;
  char new_entry;
  int value;

  /* set random generator seed */
  srand48(seed);

  /* generate values for all the cells; entries related to the
     primary suppression cells will be updated later */
  for(i=0;i<tab->m;i++) {
    for(j=0;j<tab->n;j++) {
      for(k=0;k<tab->l;k++) {
	value= int_uniform(1,1250); /* 80% of the values is in 1..1000 */
	if (value>1000) value=0; /* 20% of the values greater than 1000 are set to 0 */
	put_cell3D(tab,i,j,k,value);
      }
    }
  }
  
  /* generate primary suppression cells */
  for(i=0;i<tab->p;i++){
    new_entry= FALSE;
    while (!new_entry) { 
      r= int_uniform(0,tab->m-1);
      c= int_uniform(0,tab->n-1);
      l= int_uniform(0,tab->l-1);
      j=0;
      new_entry= TRUE;
      while((j<i)&&(new_entry)) {
	new_entry= !((r==tab->p_m[j])&&(c==tab->p_n[j])&&(l==tab->p_l[j]));
	j++;
      }
    } 
    tab->p_o[i]= i;
    tab->p_m[i]= r;
    tab->p_n[i]= c;
    tab->p_l[i]= l;
    value= int_uniform(1,1000);
    put_cell3D(tab,r,c,l,value);
    tab->p_lpl[i]= (int)ceil(pl/100.0*(double)value);;
    tab->p_upl[i]= (int)ceil(pu/100.0*(double)value);
   }
  
  /* compute marginal values */
  compute_marginals(tab);
}

void generate_table_2(TABLE3D *tab, double pl, double pu, long seed)
{
  int i,j,k,v,r,c,l;
  char new_entry;
  int value;

  /* set random generator seed */
  srand48(seed);

  /* generate values for all the cells; entries related to the
     primary suppression cells will be updated later */
  for(i=0;i<tab->m;i++) {
    for(j=0;j<tab->n;j++) {
      for(k=0;k<tab->l;k++) {
	value= int_uniform(5,501);
	if (value==501) value=0;
	put_cell3D(tab,i,j,k,value);
      }
    }
  }

  /* generate primary suppression cells */
  for(i=0;i<tab->p;i++){
    new_entry= FALSE;
    while (!new_entry) { 
      r= int_uniform(0,tab->m-1);
      c= int_uniform(0,tab->n-1);
      l= int_uniform(0,tab->l-1);
      j=0;
      new_entry= TRUE;
      while((j<i)&&(new_entry)) {
	new_entry= !((r==tab->p_m[j])&&(c==tab->p_n[j])&&(l==tab->p_l[j]));
	j++;
      }
    } 
    tab->p_o[i]= i;
    tab->p_m[i]= r;
    tab->p_n[i]= c;
    tab->p_l[i]= l;
    value= int_uniform(1,4);
    put_cell3D(tab,r,c,l,value);
    tab->p_lpl[i]= (int)ceil(pl/100.0*(double)value);;
    tab->p_upl[i]= (int)ceil(pu/100.0*(double)value);
  }
  
  /* compute marginal values */
  compute_marginals(tab);
}

/* randomly generate a value of the integer uniform distribution 
   {min,...,max} */
int int_uniform(int min, int max)
{
#define MAX_RAND48 2147483648.0
  return( (int) ((double) min + (double) (max-min)*((double)lrand48()) / ((double)MAX_RAND48) ));
}

/* write to standard_output in AMPL format */
void write_table_AMPL_format(TABLE3D *tab,double pl,double pu,int type,int seed)
{
  int i,j,k;

  printf("# generated with gen_CSP_instance_3D using parameters\n");
  printf("# \t\t %d %d %d %d %f %f %d %d\n\n", 
	 tab->m,tab->n,tab->l,tab->p,pl,pu,type,seed);

  printf("#m= number of rows of the table (without marginal row)\n");
  printf("param m := %d;\n\n",tab->m);

  printf("#n= number of columns of the table (without marginal column)\n");
  printf("param n := %d;\n\n",tab->n);

  printf("#l= number of levels of the table (without marginal level)\n");
  printf("param l := %d;\n\n",tab->l);

  printf("#p= number of primary suppression cells\n");
  printf("param p := %d;\n\n",tab->p);

  printf("#row, column, level, low_pl and upp_pl of each primary suppression cell\n");
  printf("param: \t\t p_r \t\t p_c \t\t p_l \t\t p_lpl \t\t p_upl:=\n");
  for(i=0;i<tab->p;i++) 
    printf("%d \t\t %d \t\t %d \t\t %d \t\t %d\t\t %d\n",i+1,tab->p_m[i]+1,
	   tab->p_n[i]+1,tab->p_l[i]+1,tab->p_lpl[i],tab->p_upl[i]);
  printf(";\n\n");
  
  printf("#(m+1)*(n+1)*(l+1) table values (including marginals)\n");
  printf("param a :=\n");
  for(i=0;i<=tab->m;i++){
    for(j=0;j<=tab->n;j++) {
      for(k=0;k<=tab->l;k++) {
	printf("%d \t\t %d\t\t %d\t\t %d\n",i+1,j+1,k+1,get_cell3D(tab,i,j,k));
      }
    }
  }
  printf(";\n\n");

}










